"""TcEx Framework Module"""
# first-party
from tcex.requests_external.external_session import ExternalSession
from tcex.requests_external.requests_external import RequestsExternal

__all__ = ['ExternalSession', 'RequestsExternal']
