"""TcEx Framework Module"""

# first-party
from tcex.api.tc.v3._gen.model._filter_model import FilterModel
from tcex.api.tc.v3._gen.model._property_model import PropertyModel

__all__ = ['FilterModel', 'PropertyModel']
